@@ -14,10 +14,15 @@ if pytest.__version__ < "3.9":
         return pathlib.Path(tmpdir)
 
 
+@pytest.fixture(scope="session", autouse=True)
+def session_home(tmpdir_factory):
+    os.environ["HOME"] = str(tmpdir_factory.mktemp("HOME"))
+
+
 @pytest.fixture(autouse=True)
-def home(mocker, tmp_path):
+def home(monkeypatch, tmp_path):
     h = tmp_path / "HOME"
-    os.environ["HOME"] = str(h)
+    monkeypatch.setenv("HOME", str(h))
     return h
 
 
@@ -31,15 +36,15 @@ def linux(tmpdir_factory):
 
 @pytest.fixture(autouse=True, scope="session")
 def fake_cross_compilers(tmpdir_factory):
-    missing = []
+    missing = {}
     for a in Architecture.supported():
         arch = Architecture(a)
-        binary = arch.makevars["CROSS_COMPILE"] + "gcc"
-        if not shutil.which(binary):
-            missing.append(binary)
+        for tool in ["gcc", "ld"]:
+            binary = arch.makevars["CROSS_COMPILE"] + tool
+            if not shutil.which(binary):
+                missing[binary] = tool
     if missing:
         testbin = tmpdir_factory.mktemp("bin")
-        gcc = "/usr/bin/gcc"
-        for p in missing:
-            os.symlink(gcc, testbin / p)
+        for p, real in missing.items():
+            os.symlink(f"/usr/bin/{real}", testbin / p)
         os.environ["PATH"] = f"{testbin}:" + os.environ["PATH"]
