@@ -300,6 +300,28 @@ def test_ctrl_c(linux, mocker, Popen):
     process.terminate.assert_called()
 
 
+def test_always_run_cleanup(linux, mocker):
+    build = Build(tree=linux)
+    mocker.patch(
+        "tuxmake.build.Build.build_all_targets", side_effect=KeyboardInterrupt()
+    )
+    with pytest.raises(KeyboardInterrupt):
+        build.run()
+    assert not build.build_dir.exists()
+
+
+def test_existing_build_dir(linux, home):
+    (home / ".cache" / "tuxmake" / "builds" / "current").mkdir(parents=True)
+    build = Build(tree=linux)
+    with pytest.raises(tuxmake.exceptions.TuxMakeException):
+        build.build_dir
+
+
+clang_version = int(
+    subprocess.check_output(["clang", "-dumpversion"], encoding="utf-8").split(".")[0]
+)
+
+
 class TestArchitecture:
     def test_x86_64(self, linux):
         result = build(tree=linux, target_arch="x86_64")
@@ -349,6 +371,13 @@ class TestArchitecture:
         result = build(tree=linux, target_arch="arc")
         assert "uImage.gz" in [str(f.name) for f in result.output_dir.glob("*")]
 
+    @pytest.mark.skipif(shutil.which("ld.lld") is None, reason="requires lld")
+    @pytest.mark.skipif(clang_version < 10, reason="requires clang 10+")
+    def test_hexagon(self, linux):
+        result = build(tree=linux, target_arch="hexagon", toolchain="clang")
+        assert result.passed
+        assert "vmlinux" in [str(f.name) for f in result.output_dir.glob("*")]
+
     def test_invalid_arch(self):
         with pytest.raises(tuxmake.exceptions.UnsupportedArchitecture):
             Architecture("foobar")
@@ -534,7 +563,9 @@ class TestRuntime:
         runtime.get_command_line.return_value = ["true"]
         build = Build(tree=linux, runtime="docker")
         build.run_cmd(["true"], interactive=True)
-        runtime.get_command_line.assert_called_with(build, ["true"], True)
+        runtime.get_command_line.assert_called_with(
+            build, ["true"], True, offline=build.offline
+        )
 
 
 class TestEnvironment:
@@ -698,19 +729,23 @@ class TestDebug:
 
 
 class TestPrepare:
-    def test_prepare_runtime_first(self, mocker):
+    def test_prepare_runtime_and_wrapper(self, mocker):
         order = []
+        mocker.patch(
+            "tuxmake.wrapper.Wrapper.prepare_host",
+            side_effect=lambda: order.append("wrapper_host"),
+        )
         mocker.patch(
             "tuxmake.runtime.NullRuntime.prepare",
             side_effect=lambda _: order.append("runtime"),
         )
         mocker.patch(
-            "tuxmake.wrapper.Wrapper.prepare",
-            side_effect=lambda _: order.append("wrapper"),
+            "tuxmake.wrapper.Wrapper.prepare_runtime",
+            side_effect=lambda _: order.append("wrapper_runtime"),
         )
         build = Build(wrapper="ccache")
         build.prepare()
-        assert order[0] == "runtime"
+        assert order == ["wrapper_host", "runtime", "wrapper_runtime"]
 
 
 class TestMissingArtifacts:
@@ -775,3 +810,50 @@ class TestKselftest:
         build = Build(tree=linux, targets=["kselftest"])
         names = [t.name for t in build.targets]
         assert names == ["config", "kselftest"]
+
+
+class TestBPFKselftest:
+    def test_bpf_kselftest_kconfig_add(self, linux, output_dir):
+        result = build(tree=linux, targets=["kselftest-bpf"], output_dir=output_dir)
+        kselftest_target = [t for t in result.targets if t.name == "kselftest-bpf"][0]
+        assert result.kconfig_add == kselftest_target.kconfig_add
+        config = output_dir / "config"
+        assert ("CONFIG_DEBUG_INFO_BTF=y") in config.read_text()
+
+
+class TestHeaders:
+    def test_basics(self, linux):
+        build = Build(tree=linux, targets=["headers"])
+        build.run()
+        assert "headers.tar.xz" in build.artifacts["headers"]
+
+
+class TestCheckEnvironment:
+    def test_basics(self, linux, Popen):
+        Popen.return_value.returncode = 0
+        build = Build(tree=linux, target_arch="arm64")
+        build.check_environment()
+        cmdline = args(Popen)
+        assert cmdline[0].endswith("/tuxmake-check-environment")
+        assert cmdline[1] == "arm64_gcc"
+        assert cmdline[2] == "aarch64-linux-gnu-"
+
+    def test_fails(self, linux, Popen):
+        Popen.return_value.returncode = 1
+        build = Build(tree=linux)
+        with pytest.raises(tuxmake.exceptions.EnvironmentCheckFailed):
+            build.check_environment()
+
+
+class TestReproducible:
+    def test_reproducible(self, linux):
+        build = Build(tree=linux)
+        assert "KBUILD_BUILD_TIMESTAMP" in build.environment
+        assert build.environment["KBUILD_BUILD_TIMESTAMP"].startswith("@")
+        assert "KBUILD_BUILD_USER" in build.environment
+        assert "KBUILD_BUILD_HOST" in build.environment
+
+    def test_reproducible_sets_constant_values(self, linux):
+        build1 = Build(tree=linux)
+        build2 = Build(tree=linux)
+        assert build1.environment == build2.environment
