@@ -11,10 +11,12 @@ from tuxmake import __version__
 from tuxmake.arch import Architecture, host_arch
 from tuxmake.toolchain import Toolchain, NoExplicitToolchain
 from tuxmake.wrapper import Wrapper
-from tuxmake.output import get_new_output_dir
+from tuxmake.output import get_default_build_dir, get_new_output_dir
 from tuxmake.target import create_target
 from tuxmake.runtime import get_runtime
 from tuxmake.metadata import MetadataExtractor
+from tuxmake.exceptions import BuildDirAlreadyExists
+from tuxmake.exceptions import EnvironmentCheckFailed
 from tuxmake.exceptions import UnrecognizedSourceTree
 from tuxmake.exceptions import UnsupportedArchitectureToolchainCombination
 from tuxmake.exceptions import UnsupportedMakeVariable
@@ -22,6 +24,7 @@ from tuxmake.log import LogParser
 from tuxmake.cmdline import CommandLine
 from tuxmake.utils import defaults
 from tuxmake.utils import quote_command_line
+from tuxmake.utils import get_directory_timestamp
 
 
 class BuildInfo:
@@ -119,7 +122,7 @@ class Build:
     - **kernel_image**: which kernel image to build, overriding the default
       kernel image name defined for the target architecture.
     - **jobs**: number of concurrent jobs to run (as in `make -j N`). `int`,
-      defaults to twice the number of available CPU cores.
+      defaults to the number of available CPU cores.
     - **runtime:** name of the runtime to use (`str`).
     - **verbose**: do a verbose build. The default is to do a silent build
       (i.e.  `make -s`).
@@ -170,15 +173,21 @@ class Build:
         self.__build_dir__ = None
         self.__build_dir_input__ = build_dir
         if self.__build_dir_input__:
-            self.auto_cleanup = False
+            self.clean_build_tree = False
         else:
-            self.auto_cleanup = auto_cleanup
+            self.clean_build_tree = True
+        self.auto_cleanup = auto_cleanup
 
         self.target_arch = target_arch and Architecture(target_arch) or host_arch
         self.toolchain = toolchain and Toolchain(toolchain) or NoExplicitToolchain()
         self.wrapper = wrapper and Wrapper(wrapper) or Wrapper("none")
 
-        self.environment = environment
+        self.timestamp = get_directory_timestamp(self.source_tree)
+        self.environment = {}
+        self.environment["KBUILD_BUILD_TIMESTAMP"] = "@" + self.timestamp
+        self.environment["KBUILD_BUILD_USER"] = "tuxmake"
+        self.environment["KBUILD_BUILD_HOST"] = "tuxmake"
+        self.environment.update(environment)
 
         self.kconfig = kconfig
         self.kconfig_add = kconfig_add
@@ -201,6 +210,7 @@ class Build:
         for t in targets:
             self.add_target(t)
         self.cleanup_targets()
+        self.extend_kconfig()
 
         if jobs:
             self.jobs = jobs
@@ -217,6 +227,8 @@ class Build:
         self.quiet = quiet
         self.debug = debug
 
+        self.offline = False
+
         self.artifacts = {"log": ["build.log"]}
         self.__logger__ = None
         self.__status__ = {}
@@ -256,6 +268,10 @@ class Build:
             t for t in self.targets if not self.__ordering_only_targets__[t.name]
         ]
 
+    def extend_kconfig(self):
+        for target in self.targets:
+            self.kconfig_add.extend(target.kconfig_add)
+
     def validate(self):
         source = Path(self.source_tree)
         files = [str(f.name) for f in source.glob("*")]
@@ -268,8 +284,9 @@ class Build:
             "# to reproduce this build locally: "
             + quote_command_line(self.cmdline.reproduce(self))
         )
+        self.wrapper.prepare_host()
         self.runtime.prepare(self)
-        self.wrapper.prepare(self)
+        self.wrapper.prepare_runtime(self)
 
     @property
     def output_dir(self):
@@ -292,8 +309,11 @@ class Build:
             self.__build_dir__ = Path(self.__build_dir_input__)
             self.__build_dir__.mkdir(parents=True, exist_ok=True)
         else:
-            self.__build_dir__ = self.output_dir / "tmp"
-            self.__build_dir__.mkdir()
+            self.__build_dir__ = get_default_build_dir()
+            try:
+                self.__build_dir__.mkdir()
+            except FileExistsError:
+                raise BuildDirAlreadyExists(self.__build_dir__)
         return self.__build_dir__
 
     def get_silent(self):
@@ -302,6 +322,14 @@ class Build:
         else:
             return ["--silent"]
 
+    @contextmanager
+    def go_offline(self):
+        self.offline = True
+        try:
+            yield
+        finally:
+            self.offline = False
+
     def run_cmd(self, origcmd, stdout=None, interactive=False):
         """
         Performs the build.
@@ -319,8 +347,10 @@ class Build:
         else:
             expect_failure = False
 
-        final_cmd = self.runtime.get_command_line(self, cmd, interactive)
-        extra_env = dict(**self.wrapper.environment, **self.environment, LANG="C.UTF-8")
+        final_cmd = self.runtime.get_command_line(
+            self, cmd, interactive, offline=self.offline
+        )
+        extra_env = dict(**self.wrapper.environment, **self.environment, LANG="C")
         env = dict(os.environ, **extra_env)
 
         logger = self.logger.stdin
@@ -374,6 +404,16 @@ class Build:
                 + ["--keep-going", f"--jobs={self.jobs}", f"O={self.build_dir}"]
                 + self.make_args
             )
+        elif part == "{tar_caf}":
+            return [
+                "tar",
+                "--sort=name",
+                "--owner=tuxmake:1000",
+                "--group=tuxmake:1000",
+                "--mtime=@" + self.timestamp,
+                "--clamp-mtime",
+                "-caf",
+            ]
         else:
             return [self.format_cmd_part(part)]
 
@@ -543,7 +583,21 @@ class Build:
         self.logger.terminate()
 
     def cleanup(self):
-        shutil.rmtree(self.build_dir, ignore_errors=True)
+        self.runtime.cleanup()
+        if self.clean_build_tree:
+            shutil.rmtree(self.build_dir, ignore_errors=True)
+
+    def check_environment(self):
+        self.runtime.prepare(self)
+        cmd = [str(self.runtime.get_check_environment_command())]
+        cmd.append(f"{self.target_arch.name}_{self.toolchain.name}")
+        cross = self.makevars.get("CROSS_COMPILE")
+        if cross:
+            cmd.append(cross)
+        result = self.run_cmd(cmd)
+        self.cleanup()
+        if not result:
+            raise EnvironmentCheckFailed()
 
     def run(self):
         """
@@ -557,22 +611,24 @@ class Build:
         with self.measure_duration("Preparation", metadata="prepare"):
             self.prepare()
 
-        with self.measure_duration("Build", metadata="build"):
-            self.build_all_targets()
+        try:
+            with self.go_offline():
+                with self.measure_duration("Build", metadata="build"):
+                    self.build_all_targets()
 
-        with self.measure_duration("Copying Artifacts", metadata="copy"):
-            for target in self.targets:
-                self.copy_artifacts(target)
+                with self.measure_duration("Copying Artifacts", metadata="copy"):
+                    for target in self.targets:
+                        self.copy_artifacts(target)
 
-        with self.measure_duration("Metadata Extraction", metadata="metadata"):
-            self.extract_metadata()
+                with self.measure_duration("Metadata Extraction", metadata="metadata"):
+                    self.extract_metadata()
+        finally:
+            with self.measure_duration("Cleanup", metadata="cleanup"):
+                self.terminate()
+                if self.auto_cleanup:
+                    self.cleanup()
 
-        with self.measure_duration("Cleanup", metadata="cleanup"):
-            self.terminate()
-            if self.auto_cleanup:
-                self.cleanup()
-
-        self.save_metadata()
+            self.save_metadata()
 
 
 def build(**kwargs):
