@@ -2,7 +2,13 @@ from tuxmake import xdg
 
 
 def get_default_output_basedir():
-    return xdg.cache_home() / "tuxmake" / "builds"
+    return xdg.cache_dir() / "builds"
+
+
+def get_default_build_dir():
+    base = get_default_output_basedir()
+    base.mkdir(parents=True, exist_ok=True)
+    return base / "current"
 
 
 def get_new_output_dir():
