@@ -3,9 +3,12 @@ import re
 import shlex
 import subprocess
 import sys
+import time
 from functools import lru_cache
+from pathlib import Path
 
 
+from tuxmake import cache
 from tuxmake.config import ConfigurableObject, split, splitmap, splitlistmap
 from tuxmake.exceptions import RuntimePreparationFailed
 from tuxmake.exceptions import InvalidRuntimeError
@@ -36,6 +39,7 @@ class Runtime(ConfigurableObject):
 
     def __init__(self):
         super().__init__(self.name)
+        self.__offline_available__ = None
 
     def __init_config__(self):
         self.toolchains = Toolchain.supported()
@@ -46,12 +50,45 @@ class Runtime(ConfigurableObject):
     def is_supported(self, arch, toolchain):
         return True
 
-    def get_command_line(self, build, cmd, interactive):
-        return cmd
+    @property
+    def offline_available(self):
+        if self.__offline_available__ is None:
+            prefix = self.get_command_prefix(False)
+            go_offline = str(self.get_go_offline_command())
+            try:
+                subprocess.check_output(
+                    [*prefix, go_offline, "true"], stderr=subprocess.STDOUT
+                )
+                self.__offline_available__ = True
+            except subprocess.CalledProcessError as exc:
+                error = exc.output.decode("utf-8").strip()
+                print(f"W: offline builds not available ({error})", file=sys.stderr)
+                self.__offline_available__ = False
+        return self.__offline_available__
+
+    def get_command_line(self, build, cmd, interactive, offline=True):
+        prefix = self.get_command_prefix(interactive)
+        if offline and self.offline_available:
+            go_offline = [str(self.get_go_offline_command())]
+        else:
+            go_offline = []
+        return [*prefix, *go_offline, *cmd]
+
+    def get_command_prefix(self, interactive):
+        return []
 
     def prepare(self, build):
         pass
 
+    def cleanup(self):
+        pass
+
+    def get_go_offline_command(self):
+        return Path(__file__).parent / self.basedir / "tuxmake-offline-build"
+
+    def get_check_environment_command(self):
+        return Path(__file__).parent / self.basedir / "tuxmake-check-environment"
+
 
 class NullRuntime(Runtime):
     name = "null"
@@ -81,6 +118,7 @@ class Image:
         target_kinds="",
         target_hosts="",
         packages="",
+        install_options="",
     ):
         self.name = name
         self.kind = kind
@@ -93,6 +131,7 @@ class Image:
         self.target_kinds = splitmap(target_kinds)
         self.target_hosts = splitlistmap(target_hosts)
         self.packages = split(packages)
+        self.install_options = install_options
         self.rebuild = rebuild
 
 
@@ -109,7 +148,6 @@ class DockerRuntime(Runtime):
         self.toolchains = split(self.config["runtime"]["toolchains"])
         for image_list, config in (
             (self.base_images, self.config["runtime"]["bases"]),
-            (self.ci_images, self.config["runtime"]["ci"]),
             (self.toolchain_images, self.toolchains),
         ):
             for entry in split(config):
@@ -117,8 +155,6 @@ class DockerRuntime(Runtime):
                     continue
                 if entry.startswith("base"):
                     group = "base"
-                elif entry.startswith("ci-"):
-                    group = "ci"
                 else:
                     group = f"{entry}_all"
                 image = Image(name=entry, group=group, **self.config[entry])
@@ -138,10 +174,11 @@ class DockerRuntime(Runtime):
         self.toolchain_images_map = {
             f"tuxmake/{image.name}": image for image in self.toolchain_images
         }
+        self.container_id = None
 
     @lru_cache(None)
     def is_supported(self, arch, toolchain):
-        image_name = toolchain.get_image(arch)
+        image_name = arch.get_image(toolchain) or toolchain.get_image(arch)
         image = self.toolchain_images_map.get(image_name)
         if image:
             return host_arch.name in image.hosts or any(
@@ -154,6 +191,7 @@ class DockerRuntime(Runtime):
         image = (
             os.getenv("TUXMAKE_IMAGE")
             or os.getenv("TUXMAKE_DOCKER_IMAGE")
+            or build.target_arch.get_image(build.toolchain)
             or build.toolchain.get_image(build.target_arch)
         )
         registry = os.getenv("TUXMAKE_IMAGE_REGISTRY", DEFAULT_CONTAINER_REGISTRY)
@@ -170,24 +208,28 @@ class DockerRuntime(Runtime):
     def prepare(self, build):
         super().prepare(build)
         try:
-            self.do_prepare(build)
+            self.prepare_image(build)
+            self.start_container(build)
         except subprocess.CalledProcessError:
             raise RuntimePreparationFailed(
                 self.prepare_failed_msg.format(image=self.get_image(build))
             )
 
-    def do_prepare(self, build):
-        subprocess.check_call([self.command, "pull", self.get_image(build)])
+    def prepare_image(self, build):
+        pull = [self.command, "pull", self.get_image(build)]
+        last_pull = cache.get(pull)
+        now = time.time()
+        if last_pull:
+            a_day_ago = now - (24 * 60 * 60)
+            if last_pull > a_day_ago:
+                return
+        subprocess.check_call(pull)
+        cache.set(pull, time.time())
 
-    def get_command_line(self, build, cmd, interactive):
+    def start_container(self, build):
         source_tree = os.path.abspath(build.source_tree)
         build_dir = os.path.abspath(build.build_dir)
 
-        if interactive:
-            interactive_opts = ["--interactive", "--tty"]
-        else:
-            interactive_opts = []
-
         wrapper = build.wrapper
         wrapper_opts = []
         if wrapper.path:
@@ -201,31 +243,72 @@ class DockerRuntime(Runtime):
                 v = path
             wrapper_opts.append(f"--env={k}={v}")
 
+        orig_go_offline = super().get_go_offline_command()
+        go_offline = self.get_go_offline_command()
+        go_offline_volume = self.volume(orig_go_offline, f"/usr/local/bin/{go_offline}")
+        orig_check_environment = super().get_check_environment_command()
+        check_environment = self.get_check_environment_command()
+        check_environment_volume = self.volume(
+            orig_check_environment, f"/usr/local/bin/{check_environment}"
+        )
+
         env = (f"--env={k}={v}" for k, v in build.environment.items())
         user_opts = self.get_user_opts()
         extra_opts = self.__get_extra_opts__()
-        return [
+        cmd = [
             self.command,
             "run",
             "--rm",
             "--init",
-            *interactive_opts,
+            "--detach",
             *wrapper_opts,
             "--env=KBUILD_BUILD_USER=tuxmake",
             *env,
             *user_opts,
             self.volume(source_tree, source_tree),
             self.volume(build_dir, build_dir),
+            go_offline_volume,
+            check_environment_volume,
             f"--workdir={source_tree}",
+            *self.get_logging_opts(),
             *extra_opts,
             self.get_image(build),
-        ] + cmd
+            "sleep",
+            "1d",
+        ]
+        build.log_debug(f"Starting container: {cmd}")
+        self.container_id = self.spawn_container(cmd)
+        build.log_debug(f"Container ID: {self.container_id}")
+
+    def spawn_container(self, cmd):
+        return subprocess.check_output(cmd).strip().decode("utf-8")
+
+    def get_command_prefix(self, interactive):
+        if interactive:
+            interactive_opts = ["--interactive", "--tty"]
+        else:
+            interactive_opts = []
+        return [self.command, "exec", *interactive_opts, self.container_id]
+
+    def get_go_offline_command(self):
+        return super().get_go_offline_command().name
+
+    def get_check_environment_command(self):
+        return super().get_check_environment_command().name
+
+    def cleanup(self):
+        subprocess.check_call(
+            [self.command, "stop", self.container_id], stdout=subprocess.DEVNULL
+        )
 
     def get_user_opts(self):
         uid = os.getuid()
         gid = os.getgid()
         return [f"--user={uid}:{gid}"]
 
+    def get_logging_opts(self):
+        return []
+
     def volume(self, source, target):
         return f"--volume={source}:{target}"
 
@@ -242,6 +325,9 @@ class PodmanRuntime(DockerRuntime):
     def get_user_opts(self):
         return []
 
+    def get_logging_opts(self):
+        return ["--log-level=ERROR"]
+
     def volume(self, source, target):
         return super().volume(source, target) + ":z"
 
@@ -249,7 +335,7 @@ class PodmanRuntime(DockerRuntime):
 class LocalMixin:
     prepare_failed_msg = "image {image} not found locally"
 
-    def do_prepare(self, build):
+    def prepare_image(self, build):
         subprocess.check_call(
             [self.command, "image", "inspect", self.get_image(build)],
             stdout=subprocess.DEVNULL,
